/*
 * Copyright (C) 2013 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#include "config.h"
#include "DFGJITFinalizer.h"

#if ENABLE(DFG_JIT)

#include "CodeBlock.h"
#include "DFGCommon.h"
#include "DFGPlan.h"

namespace JSC { namespace DFG {

JITFinalizer::JITFinalizer(Plan& plan, PassRefPtr<JITCode> jitCode, PassOwnPtr<LinkBuffer> linkBuffer, MacroAssembler::Label arityCheck)
    : Finalizer(plan)
    , m_jitCode(jitCode)
    , m_linkBuffer(linkBuffer)
    , m_arityCheck(arityCheck)
{
}

JITFinalizer::~JITFinalizer()
{
}

bool JITFinalizer::finalize()
{
    finalizeCommon();
    
    m_jitCode->initializeCodeRef(m_linkBuffer->finalizeCodeWithoutDisassembly());
    m_plan.codeBlock->setJITCode(m_jitCode, MacroAssemblerCodePtr());
    
    return true;
}

bool JITFinalizer::finalizeFunction()
{
    finalizeCommon();
    
    MacroAssemblerCodePtr withArityCheck = m_linkBuffer->locationOf(m_arityCheck);
    m_jitCode->initializeCodeRef(m_linkBuffer->finalizeCodeWithoutDisassembly());
    m_plan.codeBlock->setJITCode(m_jitCode, withArityCheck);
    
    return true;
}

void JITFinalizer::finalizeCommon()
{
    if (m_plan.compilation)
        m_plan.vm.m_perBytecodeProfiler->addCompilation(m_plan.compilation);
}

} } // namespace JSC::DFG

#endif // ENABLE(DFG_JIT)

